set GAME(short) klondike
set GAME(name) "Klondike Solitaire"
set GAME(comment) "Version 2003.04.20
Copr (c) 1999-2002 Richard Breuer
richard.breuer@gmx.de / www.ricki.de"
set GAME(rules) "KLONDIKE SOLITAIRE

1. Foundations (4). (Left Top of layout)
    A. Build up in suit from Ace to King.
    B. Complete these piles to win the game.

2. Tableau (7). (Below foundations)
    A. Build down by alternate color.
    B. The topmost card is available for play on the foundations or the other tableau piles.
    C. Any built unit of cards can be moved to another tableau pile.
    D. As face down cards become exposed, they can be flipped over and are now available for play.
    E. Empty spaces can only be filled by a King, or a pile whose bottommost card is a King.

3. Stock (1). (Right Top of layout)
    A. Turn up cards three at a time from the stock.
    B. There are unlimited redeals.

4. Waste (1). (Left of stock)
    A. The topmost card is available for play on the foundations and tableau.
    
\[From www.solitaire-games.com\]"

# +-+ +-+ +-+ +-+     +-+ +-+
# |0| |1| |2| |3|     |4| |5|
# +-+ +-+ +-+ +-+     +-+ +-+
#
# +-+ +-+ +-+ +-+ +-+ +-+ +-+
# |6| |7| |8| |9| |0| |1| |2|
# +-+ +-+ +-+ +-+ +-+ +-+ +-+

array set DATA {
   STACKS	13

   0,x  40	 0,y  50	 0,xo	0	 0,yo	0
   1,x 120	 1,y  50	 1,xo	0	 1,yo	0
   2,x 200	 2,y  50	 2,xo	0	 2,yo	0
   3,x 280	 3,y  50	 3,xo	0	 3,yo	0

   4,x 440	 4,y  50 	 4,xo	0	 4,yo	0
   5,x 520	 5,y  50 	 5,xo	0	 5,yo	0

   6,x  40	 6,y 200 	 6,xo	0	 6,yo	20
   7,x 120	 7,y 200 	 7,xo	0	 7,yo	20
   8,x 200	 8,y 200 	 8,xo	0	 8,yo	20
   9,x 280	 9,y 200	 9,xo	0	 9,yo	20
  10,x 360	10,y 200	10,xo	0	10,yo	20
  11,x 440	11,y 200	11,xo	0	11,yo	20
  12,x 520	12,y 200	12,xo	0	12,yo	20
}

proc doMove {fromStack toStack count} {
  set redo "stackMove $fromStack $toStack $count"
  set undo "stackMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doRevMove {fromStack toStack count} {
  set redo "stackRevMove $fromStack $toStack $count"
  set undo "stackRevMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doFlipMove {fromStack toStack count} {
  set redo "stackMove $fromStack $toStack $count
            stackFlip $fromStack 1"
  set undo "stackFlip $fromStack 1
            stackMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc doRevFlipMove {fromStack toStack count} {
  set redo "stackRevMove $fromStack $toStack $count
            stackFlip $fromStack 1"
  set undo "stackFlip $fromStack 1
            stackRevMove $toStack $fromStack $count"
  moveAdd $redo $undo
}

proc normalMove {fromStack toStack count} {
  if {[stackLength $fromStack]<$count} {return 0}
  set cl [stackGet $fromStack $count]
  if {[stackIsEmpty $toStack]} {
    # check if the last card is a king
    if {[cardValue [lindex $cl end]]!=12} {return 0}
    if {![cardsHaveFlags $cl {OPEN UPROW ALTCOL}]} {return 0}
  } else {
    lappend cl [stackGet $toStack 1]
    if {![cardsHaveFlags $cl {OPEN UPROW ALTCOL}]} {return 0}
  }

  # eventually flip
  set cl [stackGet $fromStack [expr $count+1]]
  set cl [lindex $cl end]
  if {[cardFlipped $cl]==1} {
    doRevFlipMove $fromStack $toStack $count
  } else {
    doRevMove $fromStack $toStack $count
  }
  winner
  return 1
}

proc finalMove {fromStack toStack count} {
  if {[stackLength $fromStack]<$count} {return 0}
  set cl [stackGet $fromStack $count]
  if {[stackIsEmpty $toStack]} {
    # check if the last card is a king
    if {[cardValue [lindex $cl end]]!=0} {return 0}
    if {![cardsHaveFlags $cl {OPEN UPROW PICTURE}]} {return 0}
  } else {
    lappend cl [stackGet $toStack 1]
    if {![cardsHaveFlags [revList $cl] {OPEN UPROW PICTURE}]} {return 0}
  }

  # eventually flip
  set cl [stackGet $fromStack [expr $count+1]]
  set cl [lindex $cl end]
  if {[cardFlipped $cl]==1} {
    doRevFlipMove $fromStack $toStack $count
  } else {
    doRevMove $fromStack $toStack $count
  }
  winner
  return 1
}

# THE FOLLOWING MUST BE IMPLEMENTED!!!

proc move {fromStack toStack {count 1}} {
  # never accept cards from 0..3 or 5
  if {$fromStack<4 || $fromStack==5} {return 0}
  # never accept cards to 4 or 5
  if {$toStack==4 || $toStack==5} {return 0}

  # check for the possible moves
  if {$toStack>5} {
    if {[normalMove $fromStack $toStack $count]} {autoPlay; return 1}
  } elseif {$toStack<4} {
    if {[finalMove $fromStack $toStack $count]} {autoPlay; return 1}
  }

  return 0
}

proc specialAction {stack {count 1} {type RIGHT}} {
  if {$stack==5} {
    set l [stackLength 5]
    if {$l>2} {
      set redo "stackMoveFlipped 5 4 3"
      set undo "stackMoveFlipped 4 5 3"
    } elseif {$l==2} {
      set redo "stackMoveFlipped 5 4 2"
      set undo "stackMoveFlipped 4 5 2"
    } else {
      set redo "stackMoveFlipped 5 4 1"
      set undo "stackMoveFlipped 4 5 1"
    }
    moveAdd $redo $undo
    return 1
  } elseif {$stack>3} {
    # try to find a final place for the cards
    for {set s 0} {$s<4} {incr s} {
      if {[finalMove $stack $s $count]} {autoPlay; return 1}
    }
    # try to find a normal place for the cards
    for {set s 6} {$s<13} {incr s} {
      if {$s!=$stack} {
        if {[normalMove $stack $s $count]} {autoPlay; return 1}
      }
    }
    if {$stack==4 && [stackIsEmpty 5]} {
      set redo "stackMoveFlipped 4 5 all"
      set undo "stackMoveFlipped 5 4 all"
      moveAdd $redo $undo
      return 1
    }
  }
  return 0
}

proc autoPlay {{explicitCall 0}} {
  global OPTIONS

  if {$OPTIONS(autoplay)==0} return
  if {$OPTIONS(autoplay)==1 && !$explicitCall} return

  # check for auto moves until the game is over or none is left
  set ex 0
  while {!$ex} {
    # check all final and working stacks for a move
    set ex 1
    for {set t 0} {$t<4} {incr t} {
      for {set s 6} {$s<13} {incr s} {
        if {[finalMove $s $t 1]} {
          while {[finalMove $s $t 1]} {}
          if {[winner]!=0} return
          set ex 0
        }
      }
    }
  }
}

proc finish {} {
  # return 1 for winner, -1 for loser, and 0 for game on
  for {set s 6} {$s<13} {incr s} {
    if {![stackIsEmpty $s]} {return 0}
  }
  return 1
}

proc initGame {{startSeed ""}} {
  global CARDS OPTIONS GAME widget

  set GAME(start) [clock seconds]

  set CARDS {}
  for {set i 0} {$i<52} {incr i} {lappend CARDS [expr -$i-1]}

  # OPTIONS
  # -------
  # autoplay		Choose the autoplay feature 0=never, 1=onDemand, 2=always

  set OPTIONS(autoplay) 2

  wm geometry $widget(topwindow) 580x650

  stackRemoveCards all all

  set GAME(seed) $startSeed
  if {$GAME(seed)==""} {
    set GAME(seed) [expr [clock seconds]%65536]
  }

  mix $GAME(seed)

  for {set i 0} {$i<7} {incr i} {
    set p [expr ($i*$i+$i)/2]
    stackAddCards [expr $i+6] [lrange $CARDS $p [expr $p+$i]] 0
    stackFlip [expr $i+6] 1 0
  }
  stackAddCards 5 [lrange $CARDS 28 end] 0

  stackRefresh
  moveInit

  autoPlay
}
