proc nearestStack {x y} {
  # return -1 or the index of the nearest stack to (x,y)
  global STACKS

  set min 10000000
  foreach s $STACKS(ALL) {
    set xx $STACKS($s,xpos)
    set yy $STACKS($s,ypos)
    set f [expr [llength $STACKS($s)]-1]
    set d [distPL $x $y $xx $yy \
      [expr $xx+$f*$STACKS($s,xoff)] \
      [expr $yy+$f*$STACKS($s,yoff)]]

    if {$d<$min} {
      set min $d
      set minix $s
    }
  }
  if {$min<[expr $STACKS(ALL,HEIGHT)]} {
    return $minix
  } else {
    return -1
  }
}

proc theTag {cardNo} {
  global STACKS
  return "c[card $cardNo]"
}

proc theImage {cardNo} {
  global STACKS
  if {$cardNo<0} {return [lindex $STACKS(ALL,BACKS) $STACKS(ALL,DEFBACK)]}
  return [lindex $STACKS(ALL,IMAGES) [expr $cardNo%52]]
}

proc raiseList {can list} {
  foreach c [revList $list] {$can raise $c}
}

proc setupBindings {stackNo} {
  # setup the bindings for a stack
  global STACKS

  set tmp {}
  for {set i [expr [llength $STACKS($stackNo)]-1]} {$i>=0} {incr i -1} {
    set c [theTag [lindex $STACKS($stackNo) $i]]
    lappend tmp $c

    $STACKS($stackNo,can) bind $c <1> \
      "raiseList $STACKS($stackNo,can) [list $tmp]
       itemStartDrag $STACKS($stackNo,can) %x %y"
    $STACKS($stackNo,can) bind $c <B1-Motion> \
      "itemDrag $STACKS($stackNo,can) %x %y [list $tmp]"
    $STACKS($stackNo,can) bind $c <ButtonRelease> \
      "itemStopDrag $STACKS($stackNo,can) %x %y $stackNo [list $tmp]"
    $STACKS($stackNo,can) bind $c <3> \
      "specialAction $stackNo [llength [list $tmp]] RIGHT"

    if {$STACKS($stackNo,xoff)==0 && $STACKS($stackNo,yoff)==0} return
  }
}

proc stackDisplay {stackNo {force 0}} {
  # display a stack of cards; may be empty
  global STACKS

  if {$stackNo=="all"} {
    foreach s $STACKS(ALL) {stackDisplay $s $force}
  } else {
    if {![stackExists $stackNo]} return 

    set sl [llength $STACKS($stackNo)]
    set dl [llength $STACKS($stackNo,DISPLAY)]
    if {$force=="all"} {set force $dl}

    # draw an empty stack
    if {$sl==0 && $dl==0} {
      $STACKS($stackNo,can) create rectangle \
        [expr $STACKS($stackNo,xpos)-0.5*$STACKS(ALL,WIDTH)] \
        [expr $STACKS($stackNo,ypos)-0.5*$STACKS(ALL,HEIGHT)] \
        [expr $STACKS($stackNo,xpos)+0.5*$STACKS(ALL,WIDTH)-1] \
        [expr $STACKS($stackNo,ypos)+0.5*$STACKS(ALL,HEIGHT)-1]
      return
    }

    # loop over the cards to be removed
    for {set i [expr $dl-1]} {$i>=[expr $sl-$force]} {incr i -1} {
      $STACKS($stackNo,can) delete "[theTag [lindex $STACKS($stackNo,DISPLAY) $i]]"
    }

    # loop over the cards to be drawn
    for {set i [expr $dl-$force]} {$i<$sl} {incr i} {
      $STACKS($stackNo,can) create image \
        [expr $STACKS($stackNo,xpos)+$i*$STACKS($stackNo,xoff)] \
        [expr $STACKS($stackNo,ypos)+$i*$STACKS($stackNo,yoff)] \
        -image [theImage [lindex $STACKS($stackNo) $i]] \
        -anchor center \
        -tags "[theTag [lindex $STACKS($stackNo) $i]]"
    }

    set STACKS($stackNo,DISPLAY) $STACKS($stackNo)
    update
    setupBindings $stackNo
  }
}

proc stackLength {stackNo} {
  # calc the length of a stack
  global STACKS

  if {![stackExists $stackNo]} {return 0}
  return [llength $STACKS($stackNo)]
}

proc stackIsEmpty {stackNo} {
  # determine whether a stack is empty
  global STACKS

  if {![stackExists $stackNo]} {return 1}
  return [expr [llength $STACKS($stackNo)]==0]
}

proc stackFlip {stackNo {count 1} {update 1}} {
  # flip the uppermost $count cards
  global STACKS

  if {![stackExists $stackNo]} return
  set l [llength $STACKS($stackNo)]
  if {$count=="all" || $count>$l} {set count $l}
  if {$count<1} {return 0}

  for {set i 1} {$i<=$count} {incr i} {
    set p [expr $l-$i]
    set STACKS($stackNo) [lreplace $STACKS($stackNo) $p $p \
      [expr -[lindex $STACKS($stackNo) $p]-1]]
  }
  if {$update} {stackDisplay $stackNo $count}
}

proc stackAddCards {stackNo cardNoList {update 1}} {
  # place cardNoList on top of stackNo
  global STACKS

  if {$cardNoList=={}} return
  if {![stackExists $stackNo]} return
  foreach c $cardNoList {lappend STACKS($stackNo) $c}
  if {$update} {stackDisplay $stackNo}
}

proc stackRevAddCards {stackNo cardNoList {update 1}} {
  # place cardNoList on top of stackNo in a reverse way
  global STACKS

  if {$cardNoList=={}} return
  if {![stackExists $stackNo]} return
  foreach c [revList $cardNoList] {lappend STACKS($stackNo) $c}
  if {$update} {stackDisplay $stackNo}
}

proc stackGet {stackNo {count 1}} {
  # get the uppermost $count cards w/o moving them
  # the uppermost is the first returned
  global STACKS

  if {![stackExists $stackNo]} {return {}}
  set l [llength $STACKS($stackNo)]
  if {$count=="all"} {set count $l}
  if {$count<=0} {return {}}
  if {$STACKS($stackNo)=={}} {return {}}

  set ret {}
  for {set i 0} {$i<$count} {incr i} {
    lappend ret [lindex $STACKS($stackNo) [expr $l-$i-1]]
  }
  return $ret
}

proc stackGetFlagged {stackNo flags} {
  # get as much uppermost cards as match the flags
  # the uppermost is the first returned
  global STACKS

  if {![stackExists $stackNo]} {return {}}
  set l [llength $STACKS($stackNo)]
  if {$STACKS($stackNo)=={}} {return {}}

  set ret {}
  set tmp {}
  for {set i 0} {$i<$l} {incr i} {
    lappend tmp [lindex $STACKS($stackNo) [expr $l-$i-1]]
    if {![cardsHaveFlags $tmp $flags]} {return $ret}
    set ret $tmp
  }
  return $ret
}

proc stackRemoveCards {stackNo {count 1} {update 1}} {
  # remove $count cards from the top of stackNo
  global STACKS

  if {$stackNo=="all"} {
    foreach s $STACKS(ALL) {stackRemoveCards $s $count $update}
  }

  if {![stackExists $stackNo]} return
  if {$count=="all"} {set count [llength $STACKS($stackNo)]}
  if {$count<=0} return
  set STACKS($stackNo) [lrange $STACKS($stackNo) \
    0 [expr [llength $STACKS($stackNo)]-1-$count]]
  if {$update} {stackDisplay $stackNo}
}

proc stackExists {stackNo} {
  global STACKS
  if {[lsearch -exact $STACKS(ALL) $stackNo]!=-1} {
    return 1
  } else {
    return 0
  }
}

proc stackNew {stackNo can xpos ypos {xoff 0} {yoff 0} {upd 1}} {
  # init a new stack as an empty one at graphical position $xpos $ypos in the
  # canvas widget $can. $xoff and $yoff may be set to display cards with a
  # certain offset in each direction. Offsets may be negative iff the stack
  # grows upwards (y) or to the right (x)
  global STACKS

  if {[stackExists $stackNo]} {stackRemoveCards $stackNo all}

  set STACKS($stackNo) {}
  set STACKS(ALL) [nodup [concat $STACKS(ALL) $stackNo]]
  set STACKS($stackNo,DISPLAY) {}
  foreach var {xpos ypos xoff yoff can} {
    set STACKS($stackNo,$var) [set $var]
  }
  if {$upd} {stackDisplay $stackNo}
}

proc stackRemove {stackNo} {
  # remove a stack from the list
  global STACKS

  if {[stackExists $stackNo]} {
    stackRemoveCards $stackNo all
    set tmp {}
    foreach s $STACKS(ALL) {
      if {$s!=$stackNo} {lappend tmp $s}
    }
    set STACKS(ALL) $tmp
  }
}

proc stackMove {from to {count 1} {update 1}} {
  # move $count cards from one stack to another
  # cards are moved in a one-by-one way, the first becoming the last and so on
  global STACKS

  if {![stackExists $from]} {return {}}
  if {![stackExists $to]} {return {}}
  set m [llength $STACKS($from)]
  if {$count=="all"} {set count $m}
  if {$count>$m} {set count $m}
  if {$count<=0} return

  set c [stackGet $from $count]
  stackRemoveCards $from $count 0
  stackAddCards $to $c 0
  if {$update} {
    stackDisplay $from
    stackDisplay $to
  }
}

proc stackMoveFlipped {from to {count 1} {update 1}} {
  # move $count cards from one stack to another and flips them
  # cards are moved in a one-by-one way, the first becoming the last and so on
  global STACKS

  if {![stackExists $from]} {return {}}
  if {![stackExists $to]} {return {}}
  set m [llength $STACKS($from)]
  if {$count=="all"} {set count $m}
  if {$count>$m} {set count $m}
  if {$count<=0} return

  set c [stackGet $from $count]
  stackRemoveCards $from $count 0
  stackAddCards $to [cardFlip $c] 0
  if {$update} {
    stackDisplay $from
    stackDisplay $to
  }
}

proc stackRevMove {from to {count 1} {update 1}} {
  # move $count cards from one stack to another
  # cards are moved in a reverse way, the first becoming the first and so on
  global STACKS

  if {![stackExists $from]} {return {}}
  if {![stackExists $to]} {return {}}
  set m [llength $STACKS($from)]
  if {$count=="all"} {set count $m}
  if {$count>$m} {set count $m}
  if {$count<=0} return

  set c [stackGet $from $count]
  stackRemoveCards $from $count 0
  stackRevAddCards $to $c 0
  if {$update} {
    stackDisplay $from
    stackDisplay $to
  }
}

proc stackRevMoveFlipped {from to {count 1} {update 1}} {
  # move $count cards from one stack to another, flipping them
  # cards are moved in a reverse way, the first becoming the first and so on
  global STACKS

  if {![stackExists $from]} {return {}}
  if {![stackExists $to]} {return {}}
  set m [llength $STACKS($from)]
  if {$count=="all"} {set count $m}
  if {$count>$m} {set count $m}
  if {$count<=0} return

  set c [stackGet $from $count]
  stackRemoveCards $from $count 0
  stackRevAddCards $to [cardFlip $c] 0
  if {$update} {
    stackDisplay $from
    stackDisplay $to
  }
}

proc stackRefresh {{stackNo all}} {
  # re-display all stacks
  global widget
  stackDisplay $stackNo all
  $widget(area) configure -scrollregion [$widget(area) bbox all]
}

set STACKS(ALL) {}
set STACKS(ALL,WIDTH) 71
set STACKS(ALL,HEIGHT) 96
cardReadImages ./pics STACKS(ALL,IMAGES) STACKS(ALL,BACKS)
set STACKS(ALL,DEFBACK) 0
set __stacks 1
